/*
Lgapack Version VERSION for the simulation of flow with lattice-gas automata. 
Copyright (C) 1997 D.H. Rothman and S. Zaleski. 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
in the file GNU_General_Public_License, along with this program; 
if not, write to the Free Software Foundation, Inc., 
59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

dan@segovia.mit.edu, zaleski@lmm.jussieu.fr

 Stephane Zaleski
 LMM, UPMC, Tour 66 Case 162  
 4 place Jussieu           	
 75252 PARIS Cedex 05  
 FRANCE                

 Daniel H. Rothman
 Dept of EAPS, 54-625
 MIT
 Cambridge Ma 02139
 USA
*/
/* 
  Author: S. Zaleski
  $Revision: 1.4 $ 
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h> 
#include <limits.h>
#include <sys/time.h>
#include "declarations_7.h"
#include "macros_gen.h"
#include "macros_7.h"


/* This is where parameters are defined */ 

#include "parameters_7.h"

/* 
The macro below is helpful for writing more concisely 
the propagation step on every boundary
and corner. As input it takes the coordinates 
of the site (ii, jj) and those (ii_dn, jj_dn), etc... 
of the four nearest neighbors on the hexagonal axes.
The coordinates of the four nearest neighbors are computed 
for each case, but the macro stays the same.  
*/ 

#define PROPAGATE(ii,ii_dn,ii_up,jj,jj_dn,jj_up) \
  zword  = (ii) + max_x_word_coord * (jj); \
  zword_up = (ii) + max_x_word_coord * (jj_up); \
  zword_down  = (ii) + max_x_word_coord * (jj_dn);\
  zword_next_right = (ii_up) +  max_x_word_coord * (jj);\
  zword_next_left = (ii_dn) +  max_x_word_coord * (jj); \
  a[zword] = (na[zword] >> 1) | (na[zword_next_left] << 31);\
  b[zword_up] = nb[zword];\
  c[zword_up] = (nc[zword] << 1 ) | (nc[zword_next_right] >> 31);\
  d[zword] =  (nd[zword] << 1) | (nd[zword_next_right] >>  31);\
  e[zword_down] = ne[zword]; \
  f[zword_down] = (nf[zword] >> 1) |  (nf[zword_next_left] <<  31); 

/*
  some machines have different sizes for integers, so we compute the 
  length of the words we are going to use 
*/
  const int word_length = 8*sizeof(int);
  const int word_length_minus_1 = 8*sizeof(int)-1;

main()
{
  int nx,ny,nz,j,k,t,px,py,npart,nsolids;
  int zword,iword;
  int bit_order;
  int zword_up,  zword_down, zword_next_right, zword_next_left;
  int pg, force_push, force_attempts, attempt, block_len, jblock, address;
  int j2, max_y_block_coord;
  long int ratio;
  int line_xmom[NY], line_ymom[NY], line_npart[NY];
  int count_a,count_b,count_c,count_d,count_e,count_f,count_g,tstart_aver;
  const float BIG=(float) LONG_MAX + 1.;
  FILE *outfd;
  unsigned int *a, *b, *c, *d, *e, *f, *g, *s;
  unsigned int *na, *nb, *nc, *nd, *ne, *nf; 
  int  max_x_word_coord,n_words,seed;
  float dens;
  unsigned char *lat_out;
  unsigned int *latt;

/* Counting momentum for measurement of viscsotiy */ 

  double av_force, avxmom[NY], avdens[NY], multi_avxmom[NY], multi_avdens[NY]; 
  int ncount, nexp , i_exp, nrep, i_rep;
  float force_int, forcing_rate, minf, maxf; 
  double ncount_force;

/* Writing output */ 

  FILE *summaryfd;

/* Perform some basic initialisation */

  nx=NX;
  ny=NY;
  nz=nx*ny;

  /* Do minimalist input checking */

#ifdef COUNTING
  if(TMAX < TSTART_AVER) 
    {
      fprintf(stderr,"TMAX must be larger than TSTART_AVER\n");
      exit(-1);
    }
#endif

#ifdef FORCING_RATE
#ifndef SOLID_SITES
  fprintf(stderr,"Warning: forcing without solid sites will accelerate the flow without limit\n");
  exit(-1);
#endif 
#endif


  printf("word_length=%d\n",word_length); 
  n_words=nz/word_length;
  max_x_word_coord=nx/word_length;
  dens=DENSITY;
  
/* 
    We may either use a fixed seed, to obtain each time the same
    pseudo random behavior of the code, 
    or initialize random number generator differently for each run 
*/

#ifdef SEED  
  seed = SEED;
#else
  seed = time(0); 
#endif

  srandom(seed);

/* Initialize the series of experiments */ 

  nexp = NEXP; 
  nrep = NREPLICAS;

#ifdef PARABOLAS
  summaryfd = fopen("summary","w"); 
#endif

/* Loop over the number of experiments */ 

  for(i_exp=nexp; i_exp >= 0  ; i_exp--)
    {

/* 
  Initialize velocity profile
*/ 

      for(j=0;j<ny;j++)
	{
	  multi_avxmom[j] = 0.;
	  multi_avdens[j] = 0.;
	}

/* Repeat the experiment nrep times */ 

      for(i_rep=0; i_rep <= nrep ; i_rep++)
	{
      /* Simple initialisation */ 
#ifdef FORCING_RATE
/*          forcing_rate = FORCING_RATE  */ 
	  minf = MINF;
	  maxf = MAXF; 
	  if(nexp > 0)
	      forcing_rate = 
        exp(log(minf) + (((float) i_exp) / nexp)*(log(maxf) - log(minf)));
	  else
	    forcing_rate=minf;

      printf("forcing_rate=%e\n",forcing_rate);
      block_len = min(word_length,ny);
      max_y_block_coord = ny/block_len;
      force_int = forcing_rate * block_len*word_length;
      force_push = (int ) ( force_int ); 
      if((force_int < 1. ) && (force_int > 0.))
	{
	  force_push = 1; 
	  printf("Low forcing rate \n");
	}
      if(forcing_rate <= 0.) 
	{
	  fprintf(stderr,"forcing_rate is <= 0 , undefine it in source to improve performance\n");
	  exit(-1);
	}
      force_attempts = 512 * (int)( force_push /((1. - dens) * dens));
      printf("force_push=%d\nforce_attempts=%d\n",force_push,force_attempts);

#endif  /*  (#ifdef FORCING_RATE) */
  
/* Initialize output and some files for graphics */
  
  if((outfd = fopen("dumpfile","w")) == NULL)
    {
      perror("Fhp7_msite: Could not open file -- dumpfile --.");
      exit(1); 
    }

#ifdef DEBUG  
  history_file_fd = fopen("m","w");
  fprintf(history_file_fd,"in=dumpfile\nn1=%d\nn2=%d\ncolfile=dpartercol\n",nx,ny);
  fclose(history_file_fd);
#endif  

  /* Test collision functions */ 
  
  test_collision(); 

  /* Reserve memory space */
  
  a  =    (unsigned int *) malloc(n_words*sizeof(unsigned  int)); 
  b  =    (unsigned int *) malloc(n_words*sizeof(unsigned  int)); 
  c  =    (unsigned int *) malloc(n_words*sizeof(unsigned  int)); 
  d  =    (unsigned int *) malloc(n_words*sizeof(unsigned int)); 
  e  =    (unsigned int *) malloc(n_words*sizeof(unsigned int)); 
  f  =    (unsigned int *) malloc(n_words*sizeof(unsigned int)); 
  g  =    (unsigned int *) malloc(n_words*sizeof(unsigned int)); 
  s  =    (unsigned int *) malloc(n_words*sizeof(unsigned int)); 
  
  na  =    (unsigned int *) malloc(n_words*sizeof(unsigned int)); 
  nb  =    (unsigned int *) malloc(n_words*sizeof(unsigned int)); 
  nc  =    (unsigned int *) malloc(n_words*sizeof(unsigned int)); 
  nd  =    (unsigned int *) malloc(n_words*sizeof(unsigned int)); 
  ne  =    (unsigned int *) malloc(n_words*sizeof(unsigned int)); 
  nf  =    (unsigned int *) malloc(n_words*sizeof(unsigned int)); 
  
  /* Notice latt is an int here */ 
  
  latt =   (unsigned int *) malloc(nz*sizeof(unsigned int)); 
  
  lat_out = (unsigned char *) malloc(nz*sizeof(char)); 
  
  /* INITIALIZE VELOCITY PROFILES and COUNTING */

  tstart_aver = TSTART_AVER;

#ifdef COUNTING
  for(j=0; j< ny; j++)
    {
      avxmom[j] = 0.;
      avdens[j] = 0.; 
      ncount = 0;
    }
#endif 
     ncount_force=0.; 
     av_force=0.; 

#ifdef SOLID_SITES
	  nsolids=0;
#else
	  nsolids= -1; 
#endif 

#ifdef BLOW_HEXAGON
     init_hexagon_fhp7(latt, a, b, c, d, e, f, g, na, nb, nc, nd, ne, nf, s, nx, ny); 
#else
     init_lat_fhp7(latt, a, b, c, d, e, f, g, na, nb, nc, nd, ne, nf, s, nx, ny, &nsolids, dens); 
#endif
  
  /* MAIN LOOP */
  
  for(t=0;t<=TMAX;t++)
    {
      /* Print and test conservation every so many steps */
	  if(t%TPRINT == 0 ) 
	    {
	      print_fhp7(latt, lat_out, a, b, c, d, e, f, g, s, nx, ny, outfd);
	      printf("t=%d\n ",t); test_cons(latt,nz,&px,&py,&npart); 
	      fflush(stdout);
	    }

/* The forcing algorithm */
#ifdef FORCING_RATE
      ratio = (long int) (BIG/(word_length*block_len));
      address = (int) (((long int) random())/ratio); 
      ncount_force += 1.; 

      for(jblock=0;jblock < max_y_block_coord ; jblock++)
	{
	  for(iword=0;iword < max_x_word_coord; iword++)
	    {
	      pg =0; k = 0;
/* pg is the number of forcing pushes  done */

	      if(force_int < 1. ) 
		{
         	  attempt = (int) (((float) random())/BIG + force_int);    
/* if attempt == 1 , perform a force push */

		  /* Debugging 
		  printf("%d",attempt); */
	          pg = 1 - attempt;
		  if(force_push != 1) 
		    {
		      printf("error on force_push !");
		      exit(-1);
		    }		
		}  
	      while(( pg < force_push ) &&  (k < force_attempts) ) 
		{ 
		  if( ( attempt != 1 ) && (force_int < 1. ) ) 
		    {
		      printf("error on attempt !");
		      exit(-1);
		    }		
/* Draw a random integer between 0 and word_length*block_len */ 
		  address = (address +1)%(word_length*block_len); 

/* Compute the coordinate of the selected word */
		  j2 = address/block_len; 
		  j = jblock*block_len + j2; 
		  bit_order = word_length -1 -(address - j2*block_len); 
		  zword = iword + max_x_word_coord * j; 

/* Test if the site is a candidate for reversing the a particle */

		  if( ((a[zword] & ~(d[zword])) & ( 1 << bit_order )) != 0 )
		    {
		      a[zword]  = a[zword] & ~(1 << bit_order); 
		      d[zword] |= (1 << bit_order); 
		      /* success */ 
		      address = (int) (((long int) random())/ratio); 
		      if(((s[zword] & (1<<bit_order)) == 0 ) && (t>tstart_aver)) 
			{
			  av_force += 2.; 
			}
		      pg++; 
		    }
		  k++;
		}
	      if(pg < force_push)
		{
		  printf(" Failed to apply forcing after pg %d, k %d \n in block iword = %d   jblock = %d \n", pg, k, iword, jblock);
		}
	    }
	}

#endif  /* FORCING_RATE */

      if(t&1) 
	{
	  for(j=0;j<ny;j++)
	    {
	      for(iword=0;iword < max_x_word_coord; iword++)
		{
		  zword = iword + max_x_word_coord * j; 
		  collision_7_right(a[zword],b[zword],c[zword],
d[zword],e[zword],f[zword],g+zword,na+zword,nb+zword,nc+zword,
nd+zword,ne+zword,nf+zword,s[zword]);
		}
	    }      
	}
      else
	{
	  for(j=0;j<ny;j++)
	    {
	      for(iword=0;iword < max_x_word_coord; iword++)
		{
		  zword = iword + max_x_word_coord * j; 
		  collision_7_left(a[zword],b[zword],c[zword],
d[zword],e[zword],f[zword],g+zword,na+zword,nb+zword,nc+zword,
nd+zword,ne+zword,nf+zword,s[zword]);
		}
	    }      
	}
	  

#ifdef INTERACT_ISO
      {
	interact(na, nb, nc, nd, ne, nf, a, b, c, d, e, f, 3, nx, ny);
	interact(a, b, c, d, e, f, na, nb, nc, nd, ne, nf, 2, nx, ny);
	if(t&1) 
	  {
	    interact32a(na, nb, nc, nd, ne, nf, a, b, c, d, e, f, nx, ny);
	    interact32b(a, b, c, d, e, f, na, nb, nc, nd, ne, nf, nx, ny);
	  }
	else
	  {
	    interact32b(na, nb, nc, nd, ne, nf, a, b, c, d, e, f, nx, ny);
	    interact32a(a, b, c, d, e, f, na, nb, nc, nd, ne, nf, nx, ny);
	  }
      }
#endif

#ifdef INTERACT_CRAS
      {
	unsigned int *tmp_ptr;  
	EXCHANGE_POINTERS(a,na)
	EXCHANGE_POINTERS(b,nb)
	EXCHANGE_POINTERS(c,nc)
	EXCHANGE_POINTERS(d,nd)
	EXCHANGE_POINTERS(e,ne)
	EXCHANGE_POINTERS(f,nf)
	r = RANGE;
	interact(a, b, c, d, e, f, na, nb, nc, nd, ne, nf, r, nx, ny);
      }
#endif



/* PROPAGATION */

      for(j=1;j < ny-1; j++)
	{
	  for(iword=1;iword < max_x_word_coord-1; iword++)
	    {
	      PROPAGATE(iword,iword-1,iword+1,j,j-1,j+1)
	    }
	      /* Treat border lines with wrap around condition  */
	    PROPAGATE(0,max_x_word_coord-1,1,j,j-1,j+1)
            PROPAGATE(max_x_word_coord-1,max_x_word_coord-2,0,j,j-1,j+1)
        }
      for(iword=1;iword < max_x_word_coord-1; iword++)
	{
	  PROPAGATE(iword,iword-1,iword+1,0,ny-1,1)
	  PROPAGATE(iword,iword-1,iword+1,ny-1,ny-2,0)
	}
	  /* Treat corners */
      PROPAGATE(0,max_x_word_coord-1,1,0,ny-1,1)
      PROPAGATE(max_x_word_coord-1,max_x_word_coord-2,0,0,ny-1,1)
      PROPAGATE(0,max_x_word_coord-1,1,ny-1,ny-2,0)
      PROPAGATE(max_x_word_coord-1,max_x_word_coord-2,0,ny-1,ny-2,0)

	  /* Measure velocity profile */

#ifdef COUNTING
if(t > tstart_aver)
	    {
	      int all_npart, all_xmom, all_ymom; 
	      
	      all_npart=all_xmom = all_ymom=0;
	      for(j=0;j<ny;j++) 
		{
		  line_xmom[j] = line_ymom[j] = line_npart[j] = 0; 
		  count_a = count_b = count_c = count_d = count_e = count_f = count_g = 0; 
		  for(iword=0;iword < max_x_word_coord; iword++)
		    {
		      zword = iword + max_x_word_coord * j; 
		      count_a +=  NUMBER_IN_WORD(a[zword]);
		      count_b +=  NUMBER_IN_WORD(b[zword]);
		      count_c +=  NUMBER_IN_WORD(c[zword]);
		      count_d +=  NUMBER_IN_WORD(d[zword]);
		      count_e +=  NUMBER_IN_WORD(e[zword]);
		      count_f +=  NUMBER_IN_WORD(f[zword]);
		      count_g +=  NUMBER_IN_WORD(g[zword]);
		    }
		  line_xmom[j]  = 2*(count_a - count_d) + count_b + count_f - count_c - count_e; 
		  line_ymom[j]  = count_b  + count_c - count_f  - count_e; 
		  line_npart[j] = count_a + count_b + count_c + count_d + count_e + count_f + count_g; 
/*
		  all_xmom  += line_xmom[j];
		  all_ymom  += line_ymom[j];
		  all_npart += line_npart[j];
*/
		  avxmom[j] += (double) (line_xmom[j]);
		  avdens[j] += (double) (line_npart[j]);
	      }
/*	      printf("all_np=%d all_xm=%d all_ym=%d\n",all_npart,all_xmom,all_ymom); 
*/
	      ncount++;
	    }
/* end if (t > tstart_aver) */ 
#endif /* COUNTING */
    }
/* End of main loop */

#ifdef PARABOLAS
/* Print some results */ 
  {
    float rho, velo[NY], b, prodvfit , fit2, nu_0, u_0 , nu_t;
    double force;
    float umax,cs;

/* Print the true density to detect any systematic error in initialisation. 
   The particles are either on non-solid sites or on 2 directions
   out of the 'border' solid site. There are 2*nx 'border' solid sites  */

    rho = ((float) npart)/(nx*ny - nsolids + (4./7.)*nx);
    printf("true density = %f\n",rho);
    printf("force cycles %0.1f\n",ncount_force);
    force = av_force/(ncount*nx*(ny-2));
    printf("average force %e\n",force);
    printf(" count %e \n",(((double) ncount)*nx*(ny-2)));

/* Compute best fit to viscosity */ 

    dens = rho/7.;
    for(j=0;j<ny;j++)
      {
	avxmom[j] = avxmom[j]/(2.*((double) ncount)*nx);
	avdens[j] = avdens[j]/(((double) ncount)*nx);
	multi_avdens[j]  += avdens[j];
	multi_avxmom[j]  += avxmom[j];
      }
    for(j=0;j<ny;j++)
      velo[j] = avxmom[j] / rho;
    b = 0.5;
    prodvfit = fit2 = 0.;
    for(j=1; j < ny-1 ; j++)
      {
	prodvfit += velo[j]*(j - b) * ( ny - 1 - b - j);
	fit2 += (j - b) * ( ny - 1 - b - j) * (j - b) * ( ny - 1 - b - j);
      }
    u_0 = prodvfit/ fit2; 
    nu_0 = - 3 * force/ (rho * 8 * u_0) ;
    nu_t = (1./28.)*(1./(dens*(1.-dens)))*(1./(1- 8.*dens*(1-dens)/7.))-1./8.;
    printf("forcing_rate %e nu_0 %e nu_t %e\n",forcing_rate,nu_0, nu_t);

    umax = u_0 *  (ny/2  - b) * ( ny/2 - 1 - b ) ;
    cs = sqrt(3./7.);
    printf("Mach = %e\n", umax/cs);
    fflush(stdout);
    fprintf(summaryfd,"%e %e %e %e %e %d \n",umax/cs,nu_0,nu_t,force,rho,i_rep);
    fflush(summaryfd);
  }
#endif
/* end of loop over replicas */

    }

#ifdef PARABOLAS
      {
    FILE *profilefd;
    char *dirname, *command, *tmp, *filename;
 
    dirname =  (char *) malloc(80*sizeof(char)); 
    command =  (char *) malloc(80*sizeof(char)); 
    tmp =  (char *) malloc(80*sizeof(char)); 
    filename =  (char *) malloc(80*sizeof(char)); 

    sprintf(dirname,"Results_Force%.1e",forcing_rate);
    tmp = strcpy(command,"mkdir ");       /* command == "mkdir "  */ 
    system(strcat(command,dirname));      /* command == mkdir dirname */
    strcpy(filename,dirname);
    strcat(filename,"/");
    strcat(filename,"profile");           /* filename ==  dirname/profile */
    
    printf("open %s \n",filename);
    profilefd = fopen(filename,"w");
    for(j=0;j<ny;j++)
      {
	multi_avxmom[j] /= (nrep+1); 
	fprintf(profilefd,"%e\n",multi_avxmom[j]);
      }
    for(j=0;j<ny;j++)
      {
	multi_avdens[j] *= 1./(nrep+1.); 
	fprintf(profilefd,"%e\n",multi_avdens[j]);
      }
    fclose(profilefd);
  }
/* end of loop over forcing rates */
#endif

    }
  fprintf(stderr,"Normal exit from Fhp7_msite ! \n");
  exit(0);
  return 1;
}

void test_cons(unsigned int *latt,int nz,int *px, int *py, int *n)
 {
  int z;
  *px= *py= *n = 0;
  for(z=0;z< nz; z++)
    {
      *px += X_MOM(latt[z]);
      *py += Y_MOM(latt[z]);
      *n  += PARTSUM(latt[z]);
    }
  fprintf(stdout," number: %d, x_mom %d, y_mom %d  \n",*n,*px,*py);
}

void errorc(unsigned int a, unsigned int b,unsigned int c,unsigned int d, unsigned int e, unsigned int f, unsigned int g, unsigned int na, unsigned int nb,unsigned int nc,unsigned int nd, unsigned int ne, unsigned int nf, unsigned int ng, char * message)
{
  fprintf(stderr,"error at %d %d %d %d %d %d %d for %s\n",a,b,c,d,e,f,g,message); 
}

