/*
Lgapack Version VERSION for the simulation of flow with lattice-gas automata. 
Copyright (C) 1997 D.H. Rothman and S. Zaleski. 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
in the file GNU_General_Public_License, along with this program; 
if not, write to the Free Software Foundation, Inc., 
59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

dan@segovia.mit.edu, zaleski@lmm.jussieu.fr

 Stephane Zaleski
 LMM, UPMC, Tour 66 Case 162  
 4 place Jussieu           	
 75252 PARIS Cedex 05  
 FRANCE                

 Daniel H. Rothman
 Dept of EAPS, 54-625
 MIT
 Cambridge Ma 02139
 USA
*/
#include "macros_gen.h"
#include "macros_7.h"
#include <stdio.h> 
#include <limits.h>
#include "declarations_7.h"

  /* INITIAL CONDITIONS */
void init_lat_fhp7(unsigned int *latt, unsigned int *a, unsigned int *b, unsigned int *c, unsigned int *d, unsigned int *e, unsigned int *f, unsigned int *g, unsigned int *na, unsigned int *nb, unsigned int *nc, unsigned int *nd, unsigned int *ne, unsigned int *nf, unsigned int *s, int nx, int ny, int *nsolids, float dens)
{
  int i,j,k,z,x,y,nz,iword, zlatt, particle, px, py, npart;
  int bit_from_left;
  int bit_order;
  int zword, n_words, max_x_word_coord, expected_npart; 
  FILE *solidfd;
  unsigned char *solid;
  float free_dens; 

  nz = nx*ny;
  n_words=nz/word_length;
  max_x_word_coord=nx/word_length;

  for(z=0;z<nz;z++)
    {
      latt[z] = 0;
    }

  if((*nsolids) == 0)
    {
        solid =  (unsigned char *) malloc(nz*sizeof(char));
  
	/* Read a file containing data about solid sites */
  
	if((solidfd = fopen("solid","r")) != NULL) 
	  {
	    i=fread((char *) solid, sizeof(char), nz*sizeof(char),solidfd);
	  }
	else
	  {
	    fprintf(stderr,"No solid site configuration found\n");
	    fprintf(stderr,"Continuing normal operation without solid sites\n");
	    for(z=0;z<nz;z++)
	      solid[z]=0;
	  }
  
 
	/* Translate to hex coordinates and place into single site words */
	/* Solid is in cartesian coordinates and latt in HEX coordinates */
	/* Count the solid sites */ 
	*nsolids = 0; 
	for(i=0;i<nx;i++)
	  for(j=0;j<ny;j++)
	    {
	      x = X_HEX_TO_CART(i,j);
	      y = Y_HEX_TO_CART(i,j); 
	      if(solid[x + nx*y]) 
		{
		  latt[i + nx*j] |= S;
		  (*nsolids)++;
		}
	    }
	printf("solid file read \n"); 
      }
  else
    *nsolids = 0;

  /* Print out the solid fraction, for info */ 
  
  printf("Solid fraction = %0.4f \n", (float) (*nsolids)/nz); 

/* We  correct here a potential systematic error in initialisation. 
   The particles are either on non-solid sites or on 2 directions
   out of the 'border' solid site, but we put them at first only on 
   non-solid sites. There are 2*nx 'border' solid sites. This results
   in an adjustment of the density as below */

   free_dens = dens*((float)(nz - *nsolids) + (4./7.)*nx)/(nz - *nsolids);

  /* Initialize the lattice with density dens  */

  for(z=0;z<nz;z++)
    {
      if(SP(latt[z]) == 0)
	{
	  for(k=0; k<7; k++)
	    {
/* 
	 The variable particle will be equal to 1 with probability free_dens 
	 and to 0 with probability 1 - free_dens 
*/ 
	      particle= (int) (((float) random())/((float) LONG_MAX + 1.) 
			       + free_dens);    
	      if(particle)
		{
		  latt[z] |= (1 << k); 
		}
	    }
	}
    }

/* We  correct here a systematic error in initialisation. 
   The particles are either on non-solid sites or on 2 directions
   out of the 'border' solid site. There are 2*nx 'border' solid sites  */

  expected_npart = (int) (7*dens*(nz-*nsolids + (4./7.)*nx));
  printf("initial conditions"); 
  test_cons(latt,nz,&px,&py,&npart); 
  add_particles(latt,nz,expected_npart-npart);
  printf("adjusted by  %d particles\n",expected_npart-npart);

  /* Translate initial conditions from the initial data format to 
     multispin format */
  
  /* First clear all bits of multispin structure */
  
  for(zword = 0 ; zword < n_words; zword++)
    {
      a[zword] = 0;
      b[zword] = 0;
      c[zword] = 0;
      d[zword] = 0;
      e[zword] = 0;
      f[zword] = 0;
      g[zword] = 0;
      s[zword] = 0;
      
      na[zword] = 0;
      nb[zword] = 0;
      nc[zword] = 0;
      nd[zword] = 0;
      ne[zword] = 0;
      nf[zword] = 0;
    }
  
  for(j=0;j < ny; j++)
    {
      for(i=0;i < nx; i++)
	{
	  zlatt = i + nx*j;
	  iword = i / word_length;
	  zword = iword + max_x_word_coord * j; 

	  /* Bit counted from left of word */

	  bit_from_left = i - iword*word_length; 

	  /* BIt order in word (counted from right) */

	  bit_order = word_length_minus_1  - bit_from_left;

	  /* Create multisite coding */

	  if(AP(latt[zlatt]))  a[zword] |= (1 << bit_order); 
	  if(BP(latt[zlatt]))  b[zword] |= (1 << bit_order); 
	  if(CP(latt[zlatt]))  c[zword] |= (1 << bit_order); 
	  if(DP(latt[zlatt]))  d[zword] |= (1 << bit_order); 
	  if(EP(latt[zlatt]))  e[zword] |= (1 << bit_order); 
	  if(FP(latt[zlatt]))  f[zword] |= (1 << bit_order); 
	  if(GP(latt[zlatt]))  g[zword] |= (1 << bit_order); 
	  if(SP(latt[zlatt]))  s[zword] |= (1 << bit_order); 
	}
    }
  
}


