#ifndef BGK_H
#define BGK_H
#include "bgk.h"
#endif

/*======================some useful macros==========================*/

#define MAP(ix,iy,iz,j)     map[(((iz)*Ny+(iy))*Nx+(ix))*19+(j)]
#define TEMPMAP(ix,iy,iz,j) temp_map[(((iz)*(Ny+2)+(iy))*(Nx+2)+(ix))*19+(j)]
#define INDEX(ix,iy,iz,j)   (((iz)*Ny+(iy))*Nx+(ix))*19+(j)  

/*==========================functions===============================*/

/********************************************************************/
void Propagate(float* lat)
{
/*------------------------------------------------------------------*
 * Propagate:
 * propagates the lattices point to the new positions
 * according their velocities and periodic boundary
 * conditions
 *------------------------------------------------------------------*/
  int i;
  static float* temp_lat=NULL;

  
  if((temp_lat=(float*) calloc(Nx*Ny*Nz*19,sizeof(float)))==NULL)
    {
      fprintf(stderr,"%s:%d allocation error\n",__FILE__,__LINE__);
      exit(1);
    }
  for(i=0;i<Nx*Ny*Nz*19;i++)
    temp_lat[i]=lat[i];
  for(i=0;i<Nx*Ny*Nz*19;i++)
      *(lat+i) = *(temp_lat+map[i]);

/*!!!! need to free memory here !!!! */
  free(temp_lat);
}


/********************************************************************/
void InitPropagate(void)
{
/*-------------------------------------------------------------------*
 * InitPropagate
 * Generates a an array map which is 
 * inside the cube the coordinates are propagated
 * for each node according their velocities, the
 * the coordinates on the faces will be propagated
 * according periodic boundary conditions.
 * Once this array of rules has been obtained, progation of the lattice
 * is then simply replacing the index by the index array map 
 *-------------------------------------------------------------------*/
  int i, ix, iy, iz, j;
  int* temp_map;

  map=(int* ) calloc(19*Nx*Ny*Nz,sizeof(int));
/*  printf("%d\t%d\n",map,19*Nx*Ny*Nz*sizeof(int)); */
  if(map == NULL)
    {
      fprintf(stderr,"%s:%d allocation error\n",__FILE__,__LINE__);
      exit(1);
    }
  temp_map=(int* ) calloc(19*(Nx+2)*(Ny+2)*(Nz+2),sizeof(int));
  if(temp_map == NULL)
    {
      fprintf(stderr,"%s:%d allocation error\n",__FILE__,__LINE__);
      exit(1);
    }

/* first propagate all points, to an extended domain*/
   for(iz = 0; iz < Nz; iz++)
     for(iy = 0; iy < Ny; iy++)
       for(ix = 0; ix < Nx; ix++)
	 for(j = 0; j < 19; j++)
	   TEMPMAP(ix+1+C(j,0),iy+1+C(j,1),iz+1+C(j,2),j)=INDEX(ix,iy,iz,j);



 /* second fill the periodic boundary conditions:
   ii) apply periodpic boundary conditions */

  Periodic(temp_map); 

  free(temp_map);
}

/********************************************************************/

void Periodic(int* temp_map)
{
  int i, ix, iy, iz, j;
  int iSizeOfLattice;

/*------------------------------------------------------------------*
 * map is equal to temp_map insize the volume Nx*Ny*Nz for each 
 * direction vector
 *------------------------------------------------------------------*/
   for(iz = 0; iz < Nz; iz++)
     for(iy = 0; iy < Ny; iy++)
       for(ix = 0; ix < Nx; ix++)
	 for(j = 0; j < 19; j++)
	   MAP(ix,iy,iz,j)=TEMPMAP(ix+1,iy+1,iz+1,j);


/*------------------------------------------------------------------*
 * at the boundaries the periodic boundary conditions apply:
 *------------------------------------------------------------------*/

/* faces without the corners */

/* planes  z = constant  */
  for ( iy = 0; iy < Ny; iy++)
    for (ix = 0; ix < Nx; ix++)
      for (j = 0; j < 18; j++)
	if(C(j,2) == 1)                                    /* z = 0 */
	  MAP(ix,iy,0,j)=TEMPMAP(ix+1,iy+1,Nz+1,j);
	else if(C(j,2) == -1)                             /* z = N3 */
	  MAP(ix,iy,Nz-1,j)=TEMPMAP(ix+1,iy+1,0,j);

/* planes y = constant */
  for ( iz = 0; iz < Nz; iz++)
    for (ix = 0; ix < Nx; ix++)
      for (j = 0; j < 18; j++)
	if(C(j,1) == 1)                               /* y = N2 */
	  MAP(ix,0,iz,j)=TEMPMAP(ix+1,Ny+1,iz+1,j);
	else if(C(j,1) == -1)                         /* y = 0 */
	  MAP(ix,Ny-1,iz,j)=TEMPMAP(ix+1,0,iz+1,j);

    
/* planes x = constant */
  for ( iz = 0; iz < Nz; iz++)
    for (iy = 0; iy < Ny; iy++)
      for (j = 0; j < 18; j++)
	if(C(j,0) == 1)                               /* x = N1 */
	  MAP(0,iy,iz,j)=TEMPMAP(Nx+1,iy+1,iz+1,j);
	else if(C(j,0) == -1)                         /* x = 0 */
	  MAP(Nx-1,iy,iz,j)=TEMPMAP(0,iy+1,iz+1,j);

/* corner points */

/* yz planes */
  for ( ix = 0; ix < Nx; ix++)
    for (j = 0; j < 18; j++ ) /* save some speed */
      {
	if ( C(j,1) == 1 && C(j,2) == 1)
	  MAP(ix,0,0,j)=TEMPMAP(ix+1,Ny+1,Nz+1,j);
	else if ( C(j,1) == 1 && C(j,2) == -1)
	  MAP(ix,0,Nz-1,j)=TEMPMAP(ix+1,Ny+1,0,j);
	else if ( C(j,1) == -1 && C(j,2) == 1)
	  MAP(ix,Ny-1,0,j)=TEMPMAP(ix+1,0,Nz+1,j);	  
	else if ( C(j,1) == -1 && C(j,2) == -1)
	  MAP(ix,Ny-1,Nz-1,j)=TEMPMAP(ix+1,0,0,j);	  
      }
    


/* xz planes */
  for ( iy = 0; iy < Ny; iy++)
    for (j = 0; j < 18; j++ )
      {
	if ( C(j,0) == 1 && C(j,2)== 1)
	  MAP(0,iy,0,j)=TEMPMAP(Nx+1,iy+1,Nz+1,j);
	else if ( C(j,0)== 1 && C(j,2) == -1 )
	  MAP(0,iy,Nz-1,j)=TEMPMAP(Nx+1,iy+1,0,j);
	else if ( C(j,0) == -1 && C(j,2) == 1 )
	  MAP(Nx-1,iy,0,j)=TEMPMAP(0,iy+1,Nz+1,j);
	else if ( C(j,0) == -1 && C(j,2) == -1 )
	  MAP(Nx-1,iy,Nz-1,j)=TEMPMAP(0,iy+1,0,j);
      }


/* xy planes */
  for ( iz = 0; iz < Nz; iz++)
    for (j = 0; j < 18; j++)
      {
	if ( C(j,0) == 1 && C(j,1) == 1)
	  MAP(0,0,iz,j)=TEMPMAP(Nx+1,Ny+1,iz+1,j);
	else if ( C(j,0) == 1 && C(j,1) == -1 )
	  MAP(0,Ny-1,iz,j)=TEMPMAP(Nx+1,0,iz+1,j);
	else if ( C(j,0) == -1 && C(j,1) == 1 )
	  MAP(Nx-1,0,iz,j)=TEMPMAP(0,Ny+1,iz+1,j);
	else if ( C(j,0) == -1 && C(j,1) == -1 )
	  MAP(Nx-1,Ny-1,iz,j)=TEMPMAP(0,0,iz+1,j);
      }
}


