#ifndef BGK_H 
#define BGK_H
#include "bgk.h"
#endif


/*#define RAND_MAX 2147483647 	*/

/**************************************************************************** 
  the following five functions are from
   Numerical Recipes, Press. et. al. 
*****************************************************************************/
#define NRANSI
#define TINY 1.0e-20
#define NR_END 1
#define FREE_ARG char*



void nrerror(char error_text[])
/* Numerical Recipes standard error handler */
{
	fprintf(stderr,"Numerical Recipes run-time error...\n");
	fprintf(stderr,"%s\n",error_text);
	fprintf(stderr,"...now exiting to system...\n");
	exit(1);
}

float *vector(long nl, long nh)
/* allocate a float vector with subscript range v[nl..nh] */
{
	float *v;

	v=(float *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(float)));
	if (!v) nrerror("allocation failure in vector()");
	return v-nl+NR_END;
}

float **matrix(long nrl, long nrh, long ncl, long nch)
/* allocate a float matrix with subscript range m[nrl..nrh][ncl..nch] */
{
	long i, nrow=nrh-nrl+1,ncol=nch-ncl+1;
	float **m;

	/* allocate pointers to rows */
	m=(float **) malloc((size_t)((nrow+NR_END)*sizeof(float*)));
	if (!m) nrerror("allocation failure 1 in matrix()");
	m += NR_END;
	m -= nrl;

	/* allocate rows and set pointers to them */
	m[nrl]=(float *) malloc((size_t)((nrow*ncol+NR_END)*sizeof(float)));
	if (!m[nrl]) nrerror("allocation failure 2 in matrix()");
	m[nrl] += NR_END;
	m[nrl] -= ncl;

	for(i=nrl+1;i<=nrh;i++) m[i]=m[i-1]+ncol;

	/* return pointer to array of pointers to rows */
	return m;
}


void free_vector(float *v, long nl, long nh)
/* free a float vector allocated with dvector() */
{
	free((FREE_ARG) (v+nl-NR_END));
}


void free_matrix(float **m, long nrl, long nrh, long ncl, long nch)
/* free a float matrix allocated by matrix() */
{
	free((FREE_ARG) (m[nrl]+ncl-NR_END));
	free((FREE_ARG) (m+nrl-NR_END));
}


void lubksb(float **a, int n, int *indx, float b[])
{
	int i,ii=0,ip,j;
	float sum;

	for (i=1;i<=n;i++) {
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii)
			for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
		else if (sum) ii=i;
		b[i]=sum;
	}
	for (i=n;i>=1;i--) {
		sum=b[i];
		for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j];
		b[i]=sum/a[i][i];
	}
}


void ludcmp(float **a, int n, int *indx, float *d)
{
	int i,imax,j,k;
	float big,dum,sum,temp;
	float *vv;

	vv=vector(1,n);
	*d=1.0;
	for (i=1;i<=n;i++) {
		big=0.0;
		for (j=1;j<=n;j++)
			if ((temp=fabs(a[i][j])) > big) big=temp;
		if (big == 0.0) nrerror("Singular matrix in routine ludcmp");
		vv[i]=1.0/big;
	}
	for (j=1;j<=n;j++) {
		for (i=1;i<j;i++) {
			sum=a[i][j];
			for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
		}
		big=0.0;
		for (i=j;i<=n;i++) {
			sum=a[i][j];
			for (k=1;k<j;k++)
				sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
			if ( (dum=vv[i]*fabs(sum)) >= big)
			  {
			    big=dum;
			    imax=i;
			  }
		}
		if (j != imax) {
			for (k=1;k<=n;k++) {
				dum=a[imax][k];
				a[imax][k]=a[j][k];
				a[j][k]=dum;
			}
			*d = -(*d);
			vv[imax]=vv[j];
		}
		indx[j]=imax;
		if (a[j][j] == 0.0) a[j][j]=TINY;
		if (j != n) {
			dum=1.0/(a[j][j]);
			for (i=j+1;i<=n;i++) a[i][j] *= dum;
		}
	}
/* for not yet knwon reason the pointer cannot freed without getting
   an error */
/*	free_vector(vv,1,n); */
}
#undef TINY
#undef NRANSI

/*--------------------------------------------------------------------------*
 * To use the numerical recipes code, a wrapper
 * has been build to hande their float pointer arithmetic.
 * Moreover, also the two routines ludcmpp and lubksb are
 * combined to solve the equation Ax=b for x 
 *--------------------------------------------------------------------------*/

void SolveAxEqualb(int n, float *A, float* b, float* x)
{
  int i,j;
  float** AA;
  float* xx;
  int* indx=NULL;
  float d;

  AA=matrix(1,n,1,n);
  xx=vector(1,n);

  for(i=0;i<n;i++)
    {
      xx[i+1] = *(b+i);
      {
	for(j=0;j<n;j++)
	  AA[i+1][j+1]=*(A+i*n+j);
      }
    }
  if((indx=(int*) malloc(sizeof(int)*n)) == NULL)
    exit(1);
  ludcmp(AA,n,indx,&d); 
  lubksb(AA,n,indx,xx);

/*--------------------------------------------------------------------------*
 * interface w/ rest of code
 *--------------------------------------------------------------------------*/ 
  for(i=0;i<n;i++)
    x[i]=xx[i+1];

  free(indx);
  free_vector(xx,1,n); 
  free_matrix(AA,1,n,1,n); 
}

